﻿// GGFront: A GHDL/GTKWave GUI Frontend
// Copyright (C) 2018-2025 Naoki FUJIEDA. New BSD License is applied.
//**********************************************************************

using Microsoft.Win32;
using System;
using System.Collections.Generic;
using System.IO;
using System.Text.RegularExpressions;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Input;

namespace GGFront
{
    // ■■ メインウィンドウ ■■
    public partial class MainWindow : Window
    {
        private MainViewModel VM;

        public MainWindow()
        {
            InitializeComponent();
            VM = new MainViewModel();
            DataContext = VM;
            foreach (int item in Util.ProcLimits)
                VM.ProcLimits.Add(item);
            foreach (int item in Util.SimLimits)
                VM.SimLimits.Add(item);
            VM.SimLimit = Util.Settings.SimLimit;
            VM.ProcLimit = Util.Settings.ProcLimit;
            UpdateHierarchy();
        }

        // 設定を表示するボタン（Setting）が押された場合
        private void ShowSettings_Click(object sender, RoutedEventArgs e)
        {
            SettingWindow win = new SettingWindow(Util.Settings);
            win.Owner = GetWindow(this);
            win.ShowDialog();
            if (win.NewSetting != null)
            {
                Util.Settings.GHDLPath = win.NewSetting.GHDLPath;
                Util.Settings.GTKWavePath = win.NewSetting.GTKWavePath;
                Util.Settings.GuessGHDLPath = win.NewSetting.GuessGHDLPath;
                Util.Settings.GuessGTKWavePath = win.NewSetting.GuessGTKWavePath;
                Util.Settings.VHDLStd = win.NewSetting.VHDLStd;
                Util.Settings.Save();
                UpdateHierarchy();
            }
        }

        // ソースを追加するボタン（Add）が押された場合
        private void AddSource_Click(object sender, RoutedEventArgs e)
        {
            OpenFileDialog dialog = new OpenFileDialog();
            dialog.Filter = "VHDL Sources (*.vhd, *.vhdl)|*.vhd;*.vhdl";
            dialog.Multiselect = true;
            if (dialog.ShowDialog() == true)
            {
                foreach (string FileName in dialog.FileNames)
                    AddSource(FileName);
                UpdateHierarchy();
            }
        }

        // ソースの一覧にファイルがドロップされた場合
        private void Sources_Drop(object sender, DragEventArgs e)
        {
            if (e.Data.GetDataPresent(DataFormats.FileDrop))
            {
                foreach (string FileName in (string[]) e.Data.GetData(DataFormats.FileDrop))
                    if (Regex.IsMatch(FileName, @"\.[vV][hH][dD][lL]?$"))
                        AddSource(FileName);
                UpdateHierarchy();
            }
        }

        private void AddSource (string FileName)
        {
            List<string> currentSources = new List<string>();
            foreach (SourceItem item in VM.SourceCollection)
                if (item.Name.Equals(FileName))
                    return;

            SourceItem newItem = new SourceItem();
            newItem.Name = FileName;
            newItem.Selected = false;
            VM.SourceCollection.Add(newItem);
        }

        // ソースを削除するボタン（Remove）またはDelキーが押された場合
        private void RemoveSource_Click(object sender, RoutedEventArgs e)
        {
            RemoveSelectedSources();
        }

        private void Sources_KeyDown(object sender, KeyEventArgs e)
        {
            if (e.Key == Key.Delete)
                RemoveSelectedSources();
        }

        private void RemoveSelectedSources()
        {
            for (int i = VM.SourceCollection.Count - 1; i >= 0; i -= 1)
                if (VM.SourceCollection[i].Selected)
                    VM.SourceCollection.RemoveAt(i);
            UpdateHierarchy();
        }

        // ソースをリセットするボタン（Reset）が押された場合
        private void Reset_Click(object sender, RoutedEventArgs e)
        {
            if (VM.SourceCollection.Count == 0)
                return;
            if (!Util.WarnAndConfirm("ソースファイル一覧がリセットされます．続けますか？"))
                return;
            VM.SourceCollection.Clear();
            UpdateHierarchy();
        }

        // トップモジュールを指定するボタン （Set as Top）が押された場合
        private void SetAsTop_Click(object sender, RoutedEventArgs e)
        {
            EntityHierarchyItem item = (EntityHierarchyItem)lstHierarchy.SelectedItem;
            if (item == null)
                return;
            Util.CurrentProject.TopModule = item.Name;
            Util.CurrentProject.GuessTopModule = false;
            UpdateHierarchy();
        }

        // 階層構造を更新するボタン（Refresh）が押された場合
        private void Refresh_Click(object sender, RoutedEventArgs e)
        {
            UpdateHierarchy();
        }

        // 階層構造を更新し，リストに表示
        private void UpdateHierarchy()
        {
            Util.CurrentProject.SourceFiles.Clear();
            foreach (SourceItem item in VM.SourceCollection)
                Util.CurrentProject.SourceFiles.Add(item.Name);

            List<EntityHierarchyItem> items = Util.CurrentProject.Hierarchy.Update();
            VM.HierarchyCollection.Clear();
            foreach (EntityHierarchyItem item in items)
                VM.HierarchyCollection.Add(item);

            if (items[0].IsValid)
            {
                string wavePath = Util.CurrentProject.WavePath;
                lblTopModule.Text = "(Waveform: " + Path.GetFileName(wavePath) + ")";
                lblVHDLStd.Text = "VHDL Version: " +
                    ((Util.Settings.VHDLStd == 1993) ? "VHDL-93" :
                     (Util.Settings.VHDLStd == 2008) ? "VHDL-2008" :
                     (Util.CurrentProject.Hierarchy.IsVHDL2008) ? "VHDL-2008 (Guessed)" : "VHDL-93 (Guessed)");
                Util.CurrentProject.UseVHDL2008 =
                    (Util.Settings.VHDLStd == 2008) ||
                    (Util.Settings.VHDLStd == 0 && Util.CurrentProject.Hierarchy.IsVHDL2008);
            }
            else
            {
                lblTopModule.Text = "";
                lblVHDLStd.Text = "VHDL Version: ";
            }
        }

        // Compile and Simulate ボタンが押された場合
        private void Compile_Click(object sender, RoutedEventArgs e)
        {
            Util.CompileAndSimulate();
        }

        // View Waveform ボタンが押された場合
        private void ViewWave_Click(object sender, RoutedEventArgs e)
        {
            // 入力をチェック
            if (!Util.Settings.Check())
                return;
            Util.Settings.Save();
            if (!Util.CurrentProject.Check())
                return;
            if (!File.Exists(Util.CurrentProject.WavePath))
            {
                Util.Warn("波形ファイルが作成されていません．");
                return;
            }
            Util.ExecTool(Util.GetGTKWavePath(), "\"" + Util.CurrentProject.WavePath + "\"", true);
        }

        // シミュレーション時間の設定が変更された場合
        private void Simlimit_SelectionChanged(object sender, SelectionChangedEventArgs e)
        {
            Util.Settings.SimLimit = VM.SimLimit;
            Util.Settings.Save();
        }

        // GHDL の実行時間制限の設定が変更された場合
        private void Proclimit_SelectionChanged(object sender, SelectionChangedEventArgs e)
        {
            Util.Settings.ProcLimit = VM.ProcLimit;
            Util.Settings.Save();
        }
    }
}
